@@ -6,6 +6,7 @@ import soco.alarms
 import pprint
 import tabulate
 import datetime
+import time
 from collections import namedtuple
 from soco.events import event_listener
 from queue import Empty
@@ -14,7 +15,10 @@ from . import sonos
 from . import speaker_info
 
 pp = pprint.PrettyPrinter(width=120)
-
+sonos_max_items = 66000
+# ToDo: Understand why a hard stop is required to exit 'wait_stopped_for'
+#       and revert the variable below. Events related?
+hard_stop = False
 
 # Error handling functions 3.7
 def error_and_exit(msg):
@@ -224,7 +228,7 @@ def no_args_one_output(speaker, action, args, soco_function, use_local_speaker_l
 
 @zero_parameters
 def list_queue(speaker, action, args, soco_function, use_local_speaker_list):
-    queue = speaker.get_queue(max_items=1000)
+    queue = speaker.get_queue(max_items=sonos_max_items)
     print_tracks(queue)
     return True
 
@@ -631,7 +635,7 @@ def list_playlist_tracks(speaker, action, args, soco_function, use_local_speaker
     if playlist:
         print("Sonos Playlist: {}".format(playlist.title))
         tracks = speaker.music_library.browse_by_idstring(
-            "sonos_playlists", playlist.item_id
+            "sonos_playlists", playlist.item_id, max_items=sonos_max_items
         )
         print_tracks(tracks)
         return True
@@ -860,12 +864,63 @@ def wait_stop(speaker, action, args, soco_function, use_local_speaker_list):
         try:
             event = sub.events.get(timeout=1.0)
             if event.variables["transport_state"] != "PLAYING":
+                logging.info(
+                    "Speaker '{}' in state '{}'".format(
+                        speaker.player_name, event.variables["transport_state"]
+                    )
+                )
                 sub.unsubscribe()
                 return True
         except Empty:
             pass
 
 
+@one_parameter
+def wait_stopped_for(speaker, action, args, soco_function, use_local_speaker_list):
+    duration = sonos.convert_to_seconds(args[0])
+    if not duration:
+        parameter_type_error(action, "Time h/m/s or HH:MM:SS")
+    logging.info("Waiting until playback stopped for {}s".format(duration))
+    sub = speaker.avTransport.subscribe(auto_renew=True)
+    while True:
+        try:
+            event = sub.events.get(timeout=1.0)
+            if event.variables["transport_state"] != "PLAYING":
+                sub.unsubscribe()
+                global hard_stop
+                hard_stop = True
+                # Poll for changes; count down reset timer
+                # ToDo: Polling is not ideal; should be redesigned using events
+                # ToDO: Use actual timestamps, not accumulated poll_intervals
+                elapsed = 0.0
+                total_elapsed_since_first_state_change = 0.0
+                poll_interval = 10.0
+                logging.info(
+                    "Checking for not PLAYING, increment = {}s".format(poll_interval)
+                )
+                while duration > elapsed:
+                    time.sleep(poll_interval)
+                    total_elapsed_since_first_state_change += poll_interval
+                    state = speaker.get_current_transport_info()[
+                        "current_transport_state"
+                    ]
+                    logging.info("Transport state = '{}'".format(state))
+                    if state != "PLAYING":
+                        elapsed += poll_interval
+                    else:
+                        elapsed = 0
+                    logging.info(
+                        "Elapsed since not 'PLAYING' = {}s, Total elapsed = {}s".format(
+                            elapsed, total_elapsed_since_first_state_change
+                        )
+                    )
+                hard_stop = False
+                return True
+        except:
+            hard_stop = False
+            pass
+
+
 @zero_parameters
 def wait_start(speaker, action, args, soco_function, use_local_speaker_list):
     sub = speaker.avTransport.subscribe(auto_renew=True)
@@ -873,6 +928,11 @@ def wait_start(speaker, action, args, soco_function, use_local_speaker_list):
         try:
             event = sub.events.get(timeout=1.0)
             if event.variables["transport_state"] == "PLAYING":
+                logging.info(
+                    "Speaker '{}' in state '{}'".format(
+                        speaker.player_name, event.variables["transport_state"]
+                    )
+                )
                 sub.unsubscribe()
                 return True
         except Empty:
@@ -1026,14 +1086,16 @@ actions = {
     "shares": SonosFunction(list_libraries, "list_library_shares"),
     "sysinfo": SonosFunction(system_info, ""),
     "sleep_at": SonosFunction(sleep_at, ""),
-    "afq": SonosFunction(add_favourite_to_queue, "add_to_queue"),
-    "add_fav_to_queue": SonosFunction(add_favourite_to_queue, "add_to_queue"),
     "add_favourite_to_queue": SonosFunction(add_favourite_to_queue, "add_to_queue"),
     "add_favorite_to_queue": SonosFunction(add_favourite_to_queue, "add_to_queue"),
+    "add_fav_to_queue": SonosFunction(add_favourite_to_queue, "add_to_queue"),
+    "afq": SonosFunction(add_favourite_to_queue, "add_to_queue"),
     "list_playlist_tracks": SonosFunction(list_playlist_tracks, "list_tracks"),
     "lpt": SonosFunction(list_playlist_tracks, "list_tracks"),
     "list_all_playlist_tracks": SonosFunction(list_all_playlist_tracks, ""),
     "lapt": SonosFunction(list_all_playlist_tracks, ""),
     "wait_stop": SonosFunction(wait_stop, ""),
     "wait_start": SonosFunction(wait_start, ""),
+    "wait_stopped_for": SonosFunction(wait_stopped_for, ""),
+    "wsf": SonosFunction(wait_stopped_for, ""),
 }
